// RedJey Graphics User Interface Library for J2ME Platform.
// Copyright (C) 2008 Yevgen Krapiva
//
// This library is free software; you can redistribute
// it and/or modify it under the terms of the
// GNU Lesser General Public License as published
// by the Free Software Foundation; either version 3.0
// of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

package com.incrediblesoft.redjey;

/**
 * @author Yevgen Krapiva<br>
 *         Homepage: http://code.google.com/p/redjey/<br>
 *         Feedback: ykrapiva@gmail.com<br>
 *         Copyright (C) 2008 Yevgen Krapiva<br>
 */

import com.incrediblesoft.redjey.event.AbstractEvent;
import com.incrediblesoft.redjey.event.MouseEvent;
import com.incrediblesoft.redjey.event.ActionEvent;
import com.incrediblesoft.redjey.event.PaintEvent;

import javax.microedition.lcdui.Graphics;
import java.util.Hashtable;
import java.util.Vector;

abstract public class Component
{
    /**
     * Layout constants
     */
    public static final int LEFT_LAYOUT    =  1;
    public static final int RIGHT_LAYOUT   =  2;
    public static final int TOP_LAYOUT     =  4;
    public static final int BOTTOM_LAYOUT  =  8;
    public static final int HCENTER_LAYOUT = 16;
    public static final int VCENTER_LAYOUT = 32;
    /**
     * Set this flag to say layout manager that this component is willing to have max available width;
     */
    public static final int H_EXPAND = 64;
    /**
     * Set this flag to say layout manager that this component is willing to have max available height;
     * Not implemented yet.
     */
    public static final int V_EXPAND = 128;

    /**
     * Parent, usually it is container
     */
    protected Component parent;
    
    /**
     * Coordinates in a parent's coordinate system (from left top edge of client area)
     */
    protected int x, y;

    /**
     * Component alignment
     */
    protected int alignment = LEFT_LAYOUT | TOP_LAYOUT;

    /**
     * Component content alignment
     */
    protected int contentAlignment = LEFT_LAYOUT | TOP_LAYOUT;

    /**
     * Width and height of component
     */
    protected int width, height;

    /**
     * Preferred fixed component size
     */
    protected Dimension preferredFixedSize;

    /**
     * Width and height of client area
     */
    protected Dimension clientSize;

    /**
     * Width and height of content area
     */
    protected Dimension contentSize;

    /**
     * Visibility flag
     */
    protected boolean isVisible;

    /**
     * Tells if this component has highlightable style
     */
    private boolean isHighlightable;

    /**
     * String tag
     */
    private String tag;

    /**
     * Component style class
     */
    protected Style[] styleSet;

    protected boolean privateStyleSet;

    /**
     * Current style
     */
    protected Style currentStyle;

    /**
     * Action listener
     */
    protected Vector actionListeners = new Vector();

    /**
     * Mouse listeners
     */
    protected Vector mouseListeners = new Vector();

    /**
     * Validation flag
     */
    protected boolean validated;

    /**
     * Predefined component states
     */
    public static final int NORMAL = 0;
    public static final int HIGHLIGHTED = 1;
    public static final int PRESSED = 1;
    public static final int DISABLED = 2;
    public static final int TOTAL_STATES = 3;

    /**
     * Component state.
     */
    protected int state = NORMAL;
    protected int prevState;

    // Variables for safe clipping and translation
    private int oldTransX, oldTransY;
    private int oldClipX, oldClipY;
    private int oldClipWidth, oldClipHeight;

    /**
     * Constructs component
     */
    public Component()
    {
        this(UIManager.DEFAULT_SET);
    }

    /**
     * Constructs component with a given style class
     * @param styleClassName style class
     */
    public Component(String styleClassName)
    {
        super();

        clientSize = new Dimension();
        contentSize = new Dimension();

        if (styleClassName == null)
            styleClassName = UIManager.DEFAULT_SET;

        setStyleClass(styleClassName);

        if (isInitiallyVisible())
            setVisible(true);
    }

    /**
     * This method is used to retrieve the appropriate style set from a hashtable of styles
     * @return Name of component
     */
    abstract public String getName();

    /**
     * Returns component's parent
     * @return parent component
     */
    public Component getParent()
    {
        return parent;
    }

    /**
     * Top containers such as Window must override this method to return different value
     * @return
     */
    protected boolean isLightweight()
    {
        return true;
    }

    /**
     * Returns true if this component can be highlighted
     * @return true/false
     */
    public boolean isHighlightable()
    {
        return isHighlightable;
    }

    /**
     * Return true if this component become visible once it is created.
     * Top containers such as Window, PopupMenu override this method.
     * @return true/false
     */
    protected boolean isInitiallyVisible()
    {
        return true;
    }

    /**
     * Returns the width of component
     * @return width
     */
    public int getWidth()
    {
        return width;
    }

    /**
     * Returns the height of component
     * @return height
     */
    public int getHeight()
    {
        return height;
    }

    /**
     * Sets the size of component
     * @param size new size
     */
    public void setSize(Dimension size)
    {
        setSize(size.width, size.height);
    }

    /**
     * Sets the size of component
     * @param width width
     * @param height height
     */
    public void setSize(int width, int height)
    {
        this.width = width;
        this.height = height;
        validate();
    }

    /**
     * Sets component location relative to parent's coordinates.
     * @param x x
     * @param y y
     */
    public void setLocation(int x, int y)
    {
        this.x = x;
        this.y = y;
    }

    public void setBounds(int x, int y, int width, int height)
    {
        setLocation(x, y);
        setSize(width, height);
    }

    public void setX(int x)
    {
        this.x = x;
    }

    public void setY(int y)
    {
        this.y = y;
    }

    public int getX()
    {
        return x;
    }

    public int getY()
    {
        return y;
    }

    /**
     * Returns X-coordinate in device's screen coordinate system
     * @return X-coordinate
     */
    public int getAbsoluteX()
    {
        return (parent != null ? parent.getAbsoluteX() + parent.getCurrentStyle().getClientOffsetX() : 0) + x;
    }

    /**
     * Returns Y-coordinate in device's screen coordinate system
     * @return Y-coordinate
     */
    public int getAbsoluteY()
    {
        return (parent != null ? parent.getAbsoluteY() + parent.getCurrentStyle().getClientOffsetY() : 0) + y;
    }

    public Rectangle getBounds()
    {
        return new Rectangle(x, y, width, height);
    }

    public Rectangle getAbsoluteRect()
    {
        return new Rectangle(getAbsoluteX(), getAbsoluteY(), width, height);
    }

    /**
     * Returns client's area dimension
     * @return Dimension
     */
    public Dimension getClientSize()
    {
        return clientSize;
    }

    /**
     * Returns the size of content area. Used to adjust content alignment
     * @return Content size
     */
    protected Dimension getContentSize()
    {
        return contentSize;
    }

    /**
     * Calculates client area dimension. Invoked by validate() method
     */
    protected void calcClientSize()
    {
        clientSize.width = getWidth() - currentStyle.getMinWidth();
        clientSize.height = getHeight() - currentStyle.getMinHeight();
    }

    /**
      * Calculates client area dimension. Invoked by validate() method
      */
     protected void calcContentSize()
     {
         contentSize.width = getWidth() - currentStyle.getMinWidth();
         contentSize.height = getHeight() - currentStyle.getMinHeight();
     }

    /**
     * Returns preferred size. Preferred size is used by Layout managers
     * @return Dimension
     */
    public Dimension getPreferredSize()
    {
        //If we have manually assigned preferred getComponentCount the return it
        if (preferredFixedSize != null) return preferredFixedSize;

        //Otherwise move along all styles and calculate the maximum preffered getComponentCount
        else if (styleSet.length > 0)
        {
            int maxWidth = 0;
            int maxHeight = 0;
            Dimension dim = null;

            for (int i=0; i<styleSet.length; i++)
            {
                if (styleSet[i] == null) continue;
                
                dim = calcPreferredSize(styleSet[i].paddingHorizontal, styleSet[i].paddingVertical, styleSet[i].font, styleSet[i].attributes);
                dim.width += styleSet[i].getMinWidth();
                dim.height += styleSet[i].getMinHeight();
                              
                maxWidth = Math.max(maxWidth, dim.width);
                maxHeight = Math.max(maxHeight, dim.height);
            }

            dim.width = maxWidth;
            dim.height = maxHeight;

            return dim;
        }

        return null;
    }

    /**
     * Calculate preferred component size depending on given style.
     * Subclasses must implement this method.
     * @param paddingHorizontal
     * @param paddingVertical
     * @param font
     * @param attributes @return The size that component wishes to have
     * @return desirable size
     */
    protected abstract Dimension calcPreferredSize(int paddingHorizontal, int paddingVertical, CustomFont font, Hashtable attributes);

    /**
     * Sets the predefined preferred dimensions.
     * @param dim Dimension
     */
    public void setPreferredSize(Dimension dim)
    {
        preferredFixedSize = dim;
    }

    /**
     * Sets component alignment hints. Used by Layout managers.
     * These hints say how to layout component.
     * @param anchor See *_LAYOUT constants
     */
    public Component setAlignment(int anchor)
    {
        alignment = anchor;
        return this;
    }

    /**
     * Returns component alignment hints
     * @return Alignment hints
     */
    public int getAlignment()
    {
        return alignment;
    }

    /**
     * Sets content alignment hints.
     * These hints say paint() method how to draw content within component client area if its size is bigger then preffered one.
     * @param anchor See *_LAYOUT constants
     */
    public void setContentAlignment(int anchor)
    {
        contentAlignment = anchor;
    }

    /**
     * Returns content alignment hints
     * @return Alignment hints
     */
    public int getContentAlignment()
    {
        return contentAlignment;
    }

    /**
     * Sets component state (NORMAL, FOCUSED, DISABLED or custom one).
     * It is also changes appearance styles.
     * @param state Component state index
     */
    public void setState(int state)
    {
        prevState = this.state;
        this.state = state;

        if (state >=0 && state < styleSet.length)
        {
            if (styleSet[state] != null)
                currentStyle = styleSet[state];
        }
    }

    /**
     * Returns current state
     * @return state
     */
    public int getState()
    {
        return state;
    }

    /**
     * Returns the current set of styles.
     * @return Set of styles
     */
    public Style[] getStyleClass()
    {
        return styleSet;
    }

    /**
     * Sets the set of styles of the corresponding class.
     * @param className The name of style class.
     */
    public void setStyleClass(String className)
    {
        this.styleSet = UIManager.getStyleClass(getName(), className);
        isHighlightable = (styleSet.length > HIGHLIGHTED && styleSet[HIGHLIGHTED] != null);
        setState(state);
        privateStyleSet = false;
    }

    public void setStyle(Style style)
    {
        setStyle(style, NORMAL);
    }

    /**
     * Sets the style that will represent some state of component
     * @param style style to set
     * @param state state to represent
     */
    public void setStyle(Style style, int state)
    {
        if (!privateStyleSet)
        {
            Style[] oldStyles = styleSet;
            styleSet = new Style[Math.max(styleSet.length, state+1)];
            for (int i=0; i<oldStyles.length; i++)
                styleSet[i] = oldStyles[i];
            privateStyleSet = true;
        }

        
        styleSet[state] = style;
        isHighlightable = (styleSet.length > HIGHLIGHTED && styleSet[HIGHLIGHTED] != null);
        setState(state);
    }

    /**
     * Returns current active style
     * @return style
     */
    public Style getCurrentStyle()
    {
        return currentStyle;
    }

    /**
     * Returns style associated with a given component's state
     * @param state component state
     * @return style
     */
    public Style getStyle(int state)
    {
        return (state < styleSet.length ? styleSet[state] : null);
    }

    /**
     * Set this component visible or not
     * @param visible true / false
     */
    public void setVisible(boolean visible)
    {
        this.isVisible = visible;
    }

    /**
     * Returns weather component is visible or not
     * @return true / false
     */
    public boolean isVisible()
    {
        return isVisible;
    }

    /**
     * Is not realized yet
     * @param keycode
     * @return
     */
    public boolean keyPressed(int keycode)
    {
        /*if (keycode == UIManager.KEY_SELECT && getAbsoluteRect().inside(UIManager.getCursorPosX(), UIManager.getCursorPosY()))
        {
            actionButtonPressed();
            return true;
        }
        else*/
            return false;
    }

    /**
     * Is not realized yet
     * @param keycode
     * @return
     */
    public boolean keyReleased(int keycode)
    {
        return false;
    }

    /**
     * Dispatches event to children
     * @param event
     */
    public void dispatchEvent(AbstractEvent event)
    {
        if (!isVisible || event.getSource() != this)
            return;
        processEvent(event);
    }

    /**
     * Processes event
     * @param e Event
     */
    protected void processEvent(AbstractEvent e)
    {        
        switch(e.getEventType())
        {
            case AbstractEvent.MOUSE_EVENT:
                MouseEvent mouseEvent = (MouseEvent) e;
                switch(mouseEvent.getID())
                {
                    case MouseEvent.MOUSE_PRESSED:
                    case MouseEvent.MOUSE_RELEASED:
                    case MouseEvent.MOUSE_CLICKED:
                    case MouseEvent.MOUSE_ENTERED:
                    case MouseEvent.MOUSE_EXITED:
                        processMouseEvent(mouseEvent);
                    break;

                    case MouseEvent.MOUSE_MOVED:
                    //case MouseEvent.MOUSE_DRAGGED:
                        processMouseMotionEvent(mouseEvent);
                        break;

                    case MouseEvent.MOUSE_SCROLL_VERTICAL:
                    case MouseEvent.MOUSE_SCROLL_HORIZONTAL:
                        processMouseScrollEvent(mouseEvent);
                        break;
                    //case MouseEvent.MOUSE_WHEEL:
                    //    processMouseWheelEvent((MouseWheelEvent)e);
                    //    break;
                }
                break;

            /*case AbstractEvent.KEY_EVENT:
                KeyEvent keyEvent = (KeyEvent) e;
                processKeyEvent(keyEvent);
                break;*/
        }
    }

    /**
     * Process mouse related events
     * @param event Event
     */
    protected void processMouseEvent(MouseEvent event)
    {
        switch(event.getID())
        {
            case MouseEvent.MOUSE_ENTERED:
                //System.out.println(""+toString()+"->Mouse entered ("+event.getX()+","+event.getY()+")");
                if (state != DISABLED && isHighlightable())
                {
                    if (state == NORMAL)
                    {
                        //System.out.println("Switching "+this+" to HIGHLIGHTED");
                        setState(HIGHLIGHTED);
                        repaint();
                    }
                }
                dispatchEventToListeners(event);
                break;

            case MouseEvent.MOUSE_EXITED:
                //System.out.println(""+toString()+"->Mouse exited");
                if (state != DISABLED && isHighlightable())
                {
                    if (state == HIGHLIGHTED || state == PRESSED)
                    {
                        //System.out.println("Switching "+this+" to NORMAL");
                        setState(NORMAL);
                        repaint();
                    }
                }
                dispatchEventToListeners(event);
                break;

            case MouseEvent.MOUSE_PRESSED:
                //System.out.println(""+toString()+"->Mouse pressed");
                if (state != DISABLED)
                {
                    setState(PRESSED);
                    repaint();
                }
                dispatchEventToListeners(event);
                break;

            case MouseEvent.MOUSE_RELEASED:
                //System.out.println(""+toString()+"->Mouse released");
                if (state != DISABLED/* && state != HIGHLIGHTED*/)
                {
                    setState(prevState);
                    repaint();
                }
                dispatchEventToListeners(event);
                break;

            case MouseEvent.MOUSE_CLICKED:
                //System.out.println(""+toString()+"->Mouse clicked");
                dispatchEventToListeners(event);
                fireActionEvent();
                break;

            default:
                return;
        }

        event.consume();
    }

    /**
     * Processes mouse motion events
     * @param event Event
     */
    protected void processMouseMotionEvent(MouseEvent event)
    {
        switch(event.getID())
        {
            case MouseEvent.MOUSE_MOVED:
                dispatchEventToListeners(event);
                break;
        }
    }

    /**
     * Processes scrolling events
     * @param event Event
     */
    protected void processMouseScrollEvent(MouseEvent event)
    {
        switch(event.getID())
        {
            case MouseEvent.MOUSE_SCROLL_VERTICAL:
            case MouseEvent.MOUSE_SCROLL_HORIZONTAL:
                dispatchEventToListeners(event);
                break;
        }
    }

    /**
     * Generates and dispatched ActionEvent to listeners
     */
    public void fireActionEvent()
    {
        dispatchEventToListeners(new ActionEvent(this));
    }

    /**
     * Returns the focus state of component
     * @return true if component is focused, false otherwise
     */
    public boolean isFocused()
    {
        return state == PRESSED;
    }

    /**
     * Reset validation flag.
     * This means that component needs to be validated again.
     */
    public void invalidate()
    {
        validated = false;
    }

    /**
     * Invalidates all component tree hierarchy     
     */
    public void invalidateTree()
    {
        if (parent != null)
            parent.invalidateTree();

        validated = false;
    }

    /**
     * Validate component
     */
    public void validate()
    {
        //System.out.println("validate -> "+this);
        
        //Calculate client area dimensions
        calcClientSize();

        //Call component's custom validation function
        validateComponent();

        //Calculate inner content getComponentCount for proper content alignment in paint() method.
        calcContentSize();

        validated = true;
    }

    /**
     * Validate component.
     * Subclasses must override this method to do custom validation procedure
     */
    protected abstract void validateComponent();

    /**
     * Adds action listener to this component.
     * @param listener action listener
     */
    public void addActionListener(ActionListener listener)
    {
        actionListeners.addElement(listener);
    }

    /**
     * Removes action listener from this component.
     * @param listener action listener
     */
    public void removeActonListener(ActionListener listener)
    {
        actionListeners.removeElement(listener);
    }

    /**
     * Adds mouse listener to this component.
     * @param listener mouse listener
     */
    public void addMouseListener(MouseListener listener)
    {
        mouseListeners.addElement(listener);
    }

    /**
     * Removes mouse listener from this component.
     * @param listener mouse listener
     */
    public void removeMouseListener(MouseListener listener)
    {
        mouseListeners.removeElement(listener);
    }

    /**
     * Displatches mouse event to listeners
     * @param mouseEvent mouse event
     */
    protected void dispatchEventToListeners(MouseEvent mouseEvent)
    {
        if (mouseListeners.size() == 0) return;

        MouseEvent event = new MouseEvent(this, mouseEvent.getID(), mouseEvent.getX(), mouseEvent.getY(), mouseEvent.getValue());
        for (int i=0; i<mouseListeners.size(); i++)
        {
            MouseListener listener = (MouseListener) mouseListeners.elementAt(i);
            switch(event.getID())
            {
                case MouseEvent.MOUSE_ENTERED:
                    listener.mouseEntered(event);
                    break;
                case MouseEvent.MOUSE_EXITED:
                    listener.mouseExited(event);
                    break;
                case MouseEvent.MOUSE_MOVED:
                    listener.mouseMoved(event);
                    break;
                case MouseEvent.MOUSE_PRESSED:
                    listener.mousePressed(event);
                    break;
                case MouseEvent.MOUSE_SCROLL_HORIZONTAL:
                case MouseEvent.MOUSE_SCROLL_VERTICAL:
                    listener.mouseScrolled(event);
                    break;
            }
        }
    }

    protected void dispatchEventToListeners(ActionEvent actionEvent)
    {
       // ActionEvent event = new ActionEvent(this);

        for (int i=0; i<actionListeners.size(); i++)
        {
            ActionListener listener = (ActionListener) actionListeners.elementAt(i);
            listener.actionPerformed(actionEvent);
        }
    }

    /**
     * Called every tick. Subclasses must override this method to do custom actions that depend on real time.
     * @param dt time in milliseconds
     */

    public void tick(int dt)
    {
        if (!isVisible)
            return;
        
        if (!validated)
        {
            //if (width <= 0 && height <= 0)
            //    setSize(preferredLayoutSize());
            //else
                validate();
        }
        
        currentStyle.tick(dt);
    }

    /**
     * Save current graphics object translation parameters
     * @param g Graphics object
     */
    protected void saveTranslate(Graphics g)
    {
        oldTransX = g.getTranslateX();
        oldTransY = g.getTranslateY();
    }

    /**
     * Save current graphics object clipping parameters
     * @param g Graphics object
     */
    protected void saveClip(Graphics g)
    {
        oldClipX = g.getClipX();
        oldClipY = g.getClipY();
        oldClipWidth = g.getClipWidth();
        oldClipHeight = g.getClipHeight();
    }

    /**
     * Restore graphics object translation parameters
     * @param g Graphics object
     */
    protected void restoreTranslate(Graphics g)
    {
        g.translate(oldTransX - g.getTranslateX(), oldTransY - g.getTranslateY());
    }

    /**
     * Restore graphics object clipping parameters
     * @param g Graphics object
     */
    protected void restoreClip(Graphics g)
    {
        g.setClip(oldClipX, oldClipY, oldClipWidth, oldClipHeight);
    }

    /**
     * Adjusts translation parameters depending on content alignment hints.
     * Called by paint() method before paiting component's content.
     * @param g Graphics object
     */
    protected void applyContentAlignment(Graphics g)
    {
        int x = 0, y = 0;

        if ((contentAlignment & LEFT_LAYOUT) != 0)
        {
        }
        else if ((contentAlignment & HCENTER_LAYOUT) != 0)
        {
            x = (clientSize.width - contentSize.width) / 2;
        }
        else if ((contentAlignment & RIGHT_LAYOUT) != 0)
        {
            x = clientSize.width - contentSize.width;
        }

        if ((contentAlignment & TOP_LAYOUT) != 0)
        {
        }
        else if ((contentAlignment & VCENTER_LAYOUT) != 0)
        {
            y = (clientSize.height - contentSize.height) / 2;
        }
        else if ((contentAlignment & BOTTOM_LAYOUT) != 0)
        {
            y = clientSize.height - contentSize.height;
        }

        if (x < 0 || x >= clientSize.width) x = 0;
        if (y < 0 || y >= clientSize.height) y = 0;
        
        g.translate(x, y);
    }

    /**
     * Paints component.
     * @param g Graphics object
     */
    public void paint(Graphics g)
    {
        if (!isVisible) return;
        
        saveTranslate(g);
        saveClip(g);

        currentStyle.paint(g, 0, 0, getWidth(), getHeight());

        //Paint component's content
        g.translate(currentStyle.getClientOffsetX(), currentStyle.getClientOffsetY());
        g.clipRect(0, 0, clientSize.width, clientSize.height);

        //Apply content alignment
        if (contentAlignment != (LEFT_LAYOUT | TOP_LAYOUT))
            applyContentAlignment(g);

        //Paint component
        paintComponent(g);

        restoreTranslate(g);
        restoreClip(g);
    }

    /**
     * Paints component. All subclasses must override this method
     * @param g graphics
     */
    abstract protected void paintComponent(Graphics g);

    /**
     * Causes application triggered repaint request
     */
    public void repaint()
    {
        repaint(0, 0, width, height);
    }

    /**
     * Dispatched repaint request
     * @param x area x
     * @param y area y
     * @param width area width
     * @param height area height
     */
    public void repaint(int x, int y, int width, int height)
    {
         if (parent != null/*this.isLightweight*/)
         {
             // Needs to be translated to parent coordinates since
             // a parent native container provides the actual repaint
             // services.  Additionally, the request is restricted to
             // the bounds of the component.
             //parent.getComponentBounds();

             Style style = parent.getCurrentStyle();

             int px = this.x + ((x < 0) ? 0 : x) + style.getClientOffsetX();
             int py = this.y + ((y < 0) ? 0 : y) + style.getClientOffsetY();
             int pwidth = (width > this.width) ? this.width : width;
             int pheight = (height > this.height) ? this.height : height;
             parent.repaint(px, py, pwidth, pheight);
         }
         else
         {
             PaintEvent e = new PaintEvent(this, new Rectangle(x, y, width, height));
             UIManager.postPaintEvent(e);
         }
    }

    /**
     * Assigns text label to component, to retrieve later from toString() method.
     * Used for debugging purposes only.
     * @param tag String label
     */
    public void setTag(String tag)
    {
        this.tag = tag;
    }

    public String toString()
    {
        return tag == null ? super.toString() : tag;
    }
}


