// RedJey Graphics User Interface Library for J2ME Platform.
// Copyright (C) 2008 Yevgen Krapiva
//
// This library is free software; you can redistribute
// it and/or modify it under the terms of the
// GNU Lesser General Public License as published
// by the Free Software Foundation; either version 3.0
// of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

package com.incrediblesoft.redjey;

/**
 * @author Yevgen Krapiva<br>
 *         Homepage: http://code.google.com/p/redjey/<br>
 *         Feedback: ykrapiva@gmail.com<br>
 *         Copyright (C) 2008 Yevgen Krapiva<br>
 */

import com.incrediblesoft.redjey.layouts.Layout;
import com.incrediblesoft.redjey.layouts.RowLayout;
import com.incrediblesoft.redjey.event.MouseEvent;
import com.incrediblesoft.redjey.event.AbstractEvent;

import javax.microedition.lcdui.Graphics;
import java.util.Vector;
import java.util.Hashtable;


public class Container extends Component
{
    /**
     * Components storage
     */
    protected Vector components;
    /**
     * Layout method that is applied to this container.
     */
    protected Layout layout;
    /**
     * Component that recieved MOUSE_ENTER message before
     */
    protected Component prevEnteredComponent;
    /**
     * Component that recieved MOUSE_PRESSED message before
     */
    //protected Component prevPressedComponent;

    /**
     * Constructs default container
     */
    public Container()
    {
        this(UIManager.DEFAULT_SET);
    }

    /**
     * Constructs container with given style class name
     *
     * @param styleClassName style class name
     */
    public Container(String styleClassName)
    {
        super(styleClassName);
        
        components = new Vector();
        setLayout(new RowLayout());
    }

    public String getName()
    {
        return "Container";
    }

    protected Dimension calcPreferredSize(int paddingHorizontal, int paddingVertical, CustomFont font, Hashtable attributes)
    {
        if (layout == null)
        {
            int maxWidth = 0, maxHeight = 0;

            for (int i = 0; i < getComponentCount(); i++)
            {
                Component c = (Component) components.elementAt(i);
                Dimension preferredSize = c.getPreferredSize();
                maxWidth = Math.max(maxWidth, c.x + preferredSize.width);
                maxHeight = Math.max(maxHeight, c.y + preferredSize.height);
            }

            return new Dimension(maxWidth, maxHeight);
        }
        else
        {
            return layout.preferredLayoutSize(this);
        }
    }

    protected void calcContentSize()
    {
        contentSize.width = 0;
        contentSize.height = 0;
        for (int i = 0; i < getComponentCount(); i++)
        {
            Component c = getComponent(i);
            contentSize.width = Math.max(contentSize.width, c.x + c.getWidth());
            contentSize.height = Math.max(contentSize.height, c.y + c.getHeight());
        }
    }


    /**
     * Returns the count of components
     *
     * @return count
     */
    public int getComponentCount()
    {
        return components.size();
    }

    /**
     * Returns component at specified index
     *
     * @param index index of component
     * @return component
     */
    public Component getComponent(int index)
    {
        if (index >= 0 && index < components.size())
            return (Component) components.elementAt(index);
        else
            return null;
    }

    /**
     * Returns component located under the x,y position in this container coordinate system.
     * Performs in-depth search.
     * 
     * @param x the x coordinate
     * @param y the y coordinate
     * @param deepSearch sets search mode. False means that only top-level
     * component will be returned, true - means perform in-septh search
     * @return null if the component does not contain the position.
     * If there is no child component at the requested point and the
     * point is within the bounds of the container the container itself
     * is returned;
     */
    public Component getComponentAt(int x, int y, boolean deepSearch)
    {
        x -= currentStyle.getClientOffsetX();
        y -= currentStyle.getClientOffsetY();

        for (int i=0; i<components.size(); i++)
        {
            Component c = (Component) components.elementAt(i);

            if (c.getBounds().inside(x, y))
            {
                if (deepSearch && (c instanceof Container))
                {
                    Component c2 = ((Container) c).getComponentAt(x - c.x, y - c.y, deepSearch);
                    if (c2 == null) c2 = c;
                    return c2;
                }
                else
                    return c;
            }
        }

        return null;
    }

    /**
     * Adds specified component to this container

     * @param c component
     * @return component
     */
    public Component add(Component c)
    {
        components.addElement(c);
        c.parent = this;
        invalidate();
        return c;
    }

    /**
     * Inserts component at specified index
     *
     * @param c component
     * @param index index the component is inserted to
     * @return component itself
     */
    public Component insertAt(Component c, int index)
    {
        components.insertElementAt(c, index);
        c.parent = this;
        invalidate();
        return c;
    }

    /**
     * Removes specified component from this container
     *
     * @param c component
     */
    public void remove(Component c)
    {
        components.removeElement(c);
        if (prevEnteredComponent == c) prevEnteredComponent = null;
        //if (prevPressedComponent == c) prevPressedComponent = null;
        invalidate();
    }

    /**
     * Removes component at specified index
     *
     * @param index index
     */
    public void remove(int index)
    {
        Component c = (Component) components.elementAt(index);
        if (c != null) remove(c);
    }

    /**
     * Removes all the components from this container
     */
    public void removeAll()
    {
        components.removeAllElements();
        prevEnteredComponent = null;
        //prevPressedComponent = null;
        invalidate();
    }

    /**
     * Returns the index of specified component within this container
     * @param c component
     * @return index of component, -1 if non was found
     */
    public int indexOf(Component c)
    {
        return components.indexOf(c);
    }

    /**
     * Sets layout to this container
     *
     * @param layout layout
     */
    public void setLayout(Layout layout)
    {
        this.layout = layout;
        invalidate();
    }

    /**
     * @return layout
     */
    public Layout getLayout()
    {
        return layout;
    }

    /**
     * Sets the size of container to preferred one.
     */
    public void pack()
    {
        setSize(getPreferredSize());
    }

    public void dispatchEvent(AbstractEvent event)
    {
        Component source = (Component) event.getSource();
        if (source == this)
            super.dispatchEvent(event);
        else if (source.parent == this)
            source.dispatchEvent(event);
    }
    
    protected void processMouseEvent(MouseEvent event)
    {
        Component target = getComponentAt(event.getX(), event.getY(), false);

        if (target != null)
        {
            event = UIManager.retargetMouseEvent(event, target);
            dispatchEvent(event);

            switch(event.getID())
            {
                /*case MouseEvent.MOUSE_PRESSED:
                    prevPressedComponent = target;
                    break;

                case MouseEvent.MOUSE_RELEASED:
                    if (prevPressedComponent == target)
                    {
                        //dispatchEvent(new MouseEvent(target, MouseEvent.MOUSE_CLICKED));
                        prevPressedComponent = null;
                    }
                    break;
                */
                case MouseEvent.MOUSE_EXITED:
                    if (prevEnteredComponent != null)
                    {
                        dispatchEvent(new MouseEvent(prevEnteredComponent, MouseEvent.MOUSE_EXITED));
                        prevEnteredComponent = null;
                    }
                    break;
            }
        }
        //else if MOUSE_EXITED target is container itself, send this event to all children
        else if (event.getID() == MouseEvent.MOUSE_EXITED)
        {
            for (int i=0; i<components.size(); i++)
                dispatchEvent(new MouseEvent((Component) components.elementAt(i), event.getID()));
        }

        if (!event.isConsumed())
            super.processMouseEvent(event);
    }

    protected void processMouseMotionEvent(MouseEvent event)
    {
        //System.out.println("MouseMotionEvent: "+toString()+" ("+event.getX()+","+event.getY()+")");
        super.processMouseMotionEvent(event);

        Component target = getComponentAt(event.getX(), event.getY(), false);
       // if (this instanceof PopupMenu)
       //     System.out.println("This "+this+"  target = "+target);
        
        if (target != null)
        {
            MouseEvent newEvent = UIManager.retargetMouseEvent(event, target);

            if (prevEnteredComponent != target)
            {
                if (prevEnteredComponent != null)
                {
                    this.dispatchEvent(new MouseEvent(prevEnteredComponent, MouseEvent.MOUSE_EXITED));
                }

                dispatchEvent(new MouseEvent(target, MouseEvent.MOUSE_ENTERED, newEvent.getX(), newEvent.getY(), 0));

                prevEnteredComponent = target;
            }

            dispatchEvent(newEvent);
            if (newEvent.isConsumed()) event.consume();
        }
        else if (prevEnteredComponent != null)
        {
            this.dispatchEvent(new MouseEvent(prevEnteredComponent, MouseEvent.MOUSE_EXITED));
            prevEnteredComponent = null;
        }


    }

    protected void processMouseScrollEvent(MouseEvent event)
    {
        super.processMouseScrollEvent(event);

        Component target = getComponentAt(event.getX(), event.getY(), false);
        if (target != null)
        {
            MouseEvent newEvent = UIManager.retargetMouseEvent(event, target);
            dispatchEvent(newEvent);
            if (newEvent.isConsumed()) event.consume();
        }
    }

    protected void applyContentAlignment(Graphics g)
    {
        // do nothing
        // container's (heavylight component) content must be controlled by layout manager
    }

    protected void validateComponent()
    {
        if (layout != null)
            layout.layoutContainer(this);
        else
            for (int i=0; i<components.size(); i++)
                ((Component) components.elementAt(i)).validate();
    }

    public void tick(int dt)
    {
        super.tick(dt);

        if (components != null)
        {
            for (int i = 0; i < components.size(); i++)
                ((Component) components.elementAt(i)).tick(dt);
        }
    }

    protected void paintComponent(Graphics g)
    {
        //g.setColor(0xFFFF0000);
        //g.drawRect(0, 0, clientSize.width-1, clientSize.height-1);
        //int num = 0;
        Rectangle clipRect = new Rectangle();
        //Current clip rect
        clipRect.x = g.getClipX();
        clipRect.y = g.getClipY();
        clipRect.width = g.getClipWidth();
        clipRect.height = g.getClipHeight();

        for (int i = 0; i < components.size(); i++)
        {
            Component c = (Component) components.elementAt(i);
             if (toString().equals("Go Home"))
             {
                 System.out.println("container -> Go Home -> paint check");
             }

            if (c.getBounds().intersects(clipRect))
            {
                //System.out.println("   "+c+" "+c.getBounds()+" is in clip rect "+clipRect+"  so paint it...");
                //System.out.println(toString()+" painting "+c.toString());
                printComponent(g, c);
                //num++;
            }
        }

       //System.out.println("Container "+toString()+" has drawn "+num+" of "+getComponentCount()+" components");
    }

    protected static void printComponent(Graphics g, Component c)
    {
        if (c != null)
        {
            int oldTransX = g.getTranslateX();
            int oldTransY = g.getTranslateY();
 
            g.translate(c.x,  c.y);
            c.paint(g);
            g.translate(oldTransX - g.getTranslateX(), oldTransY - g.getTranslateY());
        }
    }

}
