// RedJey Graphics User Interface Library for J2ME Platform.
// Copyright (C) 2008 Yevgen Krapiva
//
// This library is free software; you can redistribute
// it and/or modify it under the terms of the
// GNU Lesser General Public License as published
// by the Free Software Foundation; either version 3.0
// of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

package com.incrediblesoft.redjey;

/**
 * @author Yevgen Krapiva<br>
 *         Homepage: http://code.google.com/p/redjey/<br>
 *         Feedback: ykrapiva@gmail.com<br>
 *         Copyright (C) 2008 Yevgen Krapiva<br>
 */

import com.incrediblesoft.redjey.event.MouseEvent;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import java.util.Hashtable;

public class Menu extends MenuItem
{
    public static final int LEFT_ORIENTATION = 1;
    public static final int RIGHT_ORIENTATION = 2;
    public static final int ABOVE_ORIENTATION = 4;
    public static final int BELOW_ORIENTATION = 8;

    PopupMenu popupMenu;
    boolean isOpened;

    int popupOrientation = ABOVE_ORIENTATION;

    CustomImage arrowImage;
    
    public Menu(String title)
    {
        this(title, null, null);
    }

    public Menu(String title, CustomImage image)
    {
        this(title, image, null);
    }

    public Menu(String title, CustomImage image, String styleClassName)
    {
        super(title, image, styleClassName);
        popupMenu = new PopupMenu();
        popupMenu.parentMenu = this;
    }

    public String getName()
    {
        return "Menu";
    }

    public PopupMenu getPopupMenu()
    {
        return popupMenu;
    }

    public int getItemCount()
    {
        return popupMenu.getComponentCount();
    }

    public MenuItem getItem(int index)
    {
        return (MenuItem) popupMenu.getComponent(index);
    }

    public MenuItem add(MenuItem mi)
    {
        popupMenu.add(mi);
        return mi;
    }

    public MenuItem add(String label)
    {
        return add(new MenuItem(label));
    }

    public void remove(int index)
    {
	    popupMenu.remove(index);
    }

    public void remove(MenuItem item)
    {
        int index = popupMenu.indexOf(item);
        if (index != -1)
            remove(index);
    }

    public void removeAll()
    {
	    int nitems = getItemCount();
	    for (int i = nitems-1 ; i >= 0 ; i--)
	        remove(i);
    }

    public void processMouseEvent(MouseEvent event)
    {
        switch(event.getID())
        {
            case MouseEvent.MOUSE_CLICKED:
                if (!isOpened) open();
                else
                {
                    close();
                    setState(HIGHLIGHTED);
                }
                break;

            case MouseEvent.MOUSE_ENTERED:
                super.processMouseEvent(event);
                //if (parentMenu != null && !isOpened) open();
                break;

            default:
                super.processMouseEvent(event);
        }
    }

    public void open()
    {
        int absX = getAbsoluteX();
        int absY = getAbsoluteY();

        popupMenu.setSize(popupMenu.getPreferredSize());

        if ((popupOrientation & ABOVE_ORIENTATION) > 0)
            absY -= popupMenu.height;
        else if ((popupOrientation & BELOW_ORIENTATION) > 0)
            absY += height;
        else if ((popupOrientation & LEFT_ORIENTATION) > 0)
            absX -= popupMenu.width;
        else if ((popupOrientation & RIGHT_ORIENTATION) > 0)
            absX += width;

        popupMenu.setLocation(absX, absY);
        popupMenu.setVisible(true);

        setState(HIGHLIGHTED);

        isOpened = true;

    }

    public void close()
    {
        popupMenu.setVisible(false);
        isOpened = false;

        setState(NORMAL);
    }

    public boolean isOpened()
    {
        return isOpened;
    }

    public void setPopupOrientation(int popupOrientation)
    {
        this.popupOrientation = popupOrientation;
        popupMenu.setPopupOrientation(popupOrientation);
    }

    public void setArrowImage(CustomImage image)
    {
        this.arrowImage = image;
        invalidate();
    }

    public CustomImage getArrowImage()
    {
        return arrowImage;
    }

    public void setState(int index)
    {
        if (isOpened() && index == NORMAL)
            index = HIGHLIGHTED;
         
        super.setState(index);

        //System.out.println(toString()+" state="+index);
        
        CustomImage image = (CustomImage) currentStyle.attributes.get("arrow-image");
        if (image == null)
        {
            image = createImage();
            currentStyle.addAttribute("arrow-image", image);
            setArrowImage(image);
        }
        else
            setArrowImage(image.clone());        
    }

     private CustomImage createImage()
    {
        CustomFont font = currentStyle.getFont();


        Image image = Image.createImage(3,5);
        Graphics g = image.getGraphics();
        g.setColor(font.getColor());
        //g.drawLine(0, 0, height / 2, height / 2);
        //g.drawLine(0, height-1, height / 2, height / 2);
        g.fillTriangle(0, 0, 2, 2, 0, 4);
        return new CustomImage(image);
    }

    protected Dimension calcPreferredSize(int paddingHorizontal, int paddingVertical, CustomFont font, Hashtable attributes)
    {
        Dimension dim = super.calcPreferredSize(paddingHorizontal, paddingVertical, font, attributes);
        dim.width += arrowImage != null ? arrowImage.getWidth() + 5 : 0;
        dim.height = Math.max(dim.height, arrowImage != null ? arrowImage.getHeight(): 0);

        return dim;
    }

    protected void calcContentSize()
    {
        super.calcContentSize();
        contentSize.width += arrowImage != null ? arrowImage.getWidth() : 0;
        contentSize.height = Math.max(contentSize.height, arrowImage != null ? arrowImage.getHeight() : 0);
    }

    protected void paintComponent(Graphics g)
    {
        int prevTranslateX = g.getTranslateX();

        super.paintComponent(g);

        if (arrowImage != null && parentMenu != null)
        {
            g.translate(prevTranslateX - g.getTranslateX(), 0);
            arrowImage.paint(g, clientSize.width-1, clientSize.height >> 1, Graphics.RIGHT | Graphics.VCENTER);
        }
    }

}
